// Copyright 1999 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OWXMLProcessor.h"

#import <Foundation/Foundation.h>
#import <OmniBase/OmniBase.h>
#import <OmniFoundation/OmniFoundation.h>
#import <OWF/OWF.h>

RCS_ID("$Header: /Network/Developer/Source/CVS/OmniGroup/OWF/Processors.subproj/SGML.subproj/XML.subproj/OWXMLProcessor.m,v 1.2 1999/03/30 21:26:47 kc Exp $")

NSString *OWXMLParseException = @"OWXMLParseException";

static OWContentType *sourceContentType = nil;
static OWContentType *resultContentType = nil;

@interface OWXMLProcessor (XMLParsing)
- (void) raiseParseException: (NSString *) message;

- (BOOL) parseDocument;
- (BOOL) parseProlog;
- (BOOL) parseDTD;
- (BOOL) parseMisc;
- (BOOL) parseMiscList;
@end

@implementation OWXMLProcessor

+ (void)didLoad;
{
    sourceContentType = [[OWContentType contentTypeForString:@"text/xml"] retain];
    resultContentType = [[OWContentType contentTypeForString:@"ObjectStream/xml"] retain];

    [self registerProcessorClass: self
                 fromContentType: sourceContentType
                   toContentType: resultContentType
                            cost: 1.0];
}

- initWithPipeline:(OWPipeline *)aPipeline;
{
    if (![super initWithPipeline: aPipeline])
        return nil;

    objectStream = [[OWObjectStream alloc] init];
    [objectStream setContentType: resultContentType];

#warning Need to set the string encoding on the data cursor

    scanner = [[OWDataStreamScanner alloc] initWithCursor: dataCursor];
    [pipeline addContent: objectStream];
    [pipeline cacheContent];
    [pipeline startProcessingContent];
    
    return self;
}

- (void) dealloc;
{
    [document release];
    [objectStream release];
    [scanner release];
    [super dealloc];
}

- (void)process;
{
    [self parseDocument];
    [objectStream dataEnd];
}

- (void)processAbort;
{
    [objectStream dataAbort];
    [super processAbort];
}

@end


@implementation OWXMLProcessor (XMLParsing)

- (void) raiseParseException: (NSString *) message;
{
#warning TJW: Later we should put a user info dictionary in this that will let us display to the user the location in the stream at which the parse error occured.
    [NSException raise: OWXMLParseException
                format: @"Error parsing XML document: %@", message];
}

- (BOOL) parseDocument;
{
    [self parseProlog];
    [self parseElement];
    [self parseMiscList];
}

- (void) parseProlog;
{
    [self parseXMLDecl];
    [self parseMiscList];
    if ([self parseDTD])
        [self parseMiscList];
}

- (void) parseXMLDecl;
{
    if (![scanner scanString: @"<?xml" peek: NO])
        return NO;
    [self parseVersionInfo];
    [self parseEncodingDecl];
    [self parseSDDecl];
    [self parseSpace];

    if (![scanner scanString: @"?>" peek: NO])
        [self raiseParseException: @"foo"];
}

- (BOOL) parseDTD;
{
    
}

- (BOOL) parseMisc;
{
    
}

- (void) parseMiscList;
{
    while ([self parseMisc])
        ;
}
@end
